@@ -1,6 +1,7 @@
 # vim: tabstop=4 shiftwidth=4 softtabstop=4
 
 # Copyright 2013 IBM Corporation
+# Copyright 2015 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,9 +16,8 @@
 # limitations under the License.
 # This represents the low layer message framing portion of IPMI
 
-import atexit
 import collections
-import fcntl
+import ctypes
 import hashlib
 import hmac
 import operator
@@ -47,10 +47,7 @@ iothread = None  # the thread in which all IO will be performed
                  # the nature of things.
 iothreadready = False  # whether io thread is yet ready to work
 iothreadwaiters = []  # threads waiting for iothreadready
-ignoresockets = set()  # between 'select' firing and 'recvfrom', a socket
-                       # should be ignored
 ioqueue = collections.deque([])
-selectbreak = None
 selectdeadline = 0
 running = True
 iosockets = []  # set of iosockets that will be shared amongst Session objects
@@ -59,88 +56,80 @@ MAX_BMCS_PER_SOCKET = 64  # no more than this many BMCs will share a socket
                          # value, leading to fewer filehandles
 
 
-def _ioworker():
-    global iothreadready
-    global selectbreak
-    global selectdeadline
-    selectbreak = os.pipe()
-    fcntl.fcntl(selectbreak[0], fcntl.F_SETFL, os.O_NONBLOCK)
-    iowaiters = []
-    timeout = 300
-    iothreadready = True
-    while running:
-        while iothreadwaiters:
-            waiter = iothreadwaiters.pop()
-            waiter.set()
-        if timeout < 0:
-            timeout = 0
-        selectdeadline = _monotonic_time() + timeout
-        if ignoresockets:
-            mysockets = [selectbreak[0]]
-            for pendingsocket in iosockets:
-                if pendingsocket not in ignoresockets:
-                    mysockets.append(pendingsocket)
-        else:
-            mysockets = iosockets + [selectbreak[0]]
-        tmplist, _, _ = select.select(mysockets, (), (), timeout)
-        # pessimistically move out the deadline
-        # doing it this early (before ioqueue is evaluated)
-        # this avoids other threads making a bad assumption
-        # about not having to break into the select
-        selectdeadline = _monotonic_time() + 300
-        rdylist = []
-        for handle in tmplist:
-            if handle is selectbreak[0]:
-                try:  # flush all requests to interrupt that may be pending
-                    while True:
-                        os.read(handle, 1)
-                except OSError:
-                    # this means an EWOULDBLOCK occurred, ignore that as that
-                    # was the endgame
-                    pass
-            else:
-                ignoresockets.add(handle)
-                rdylist.append(handle)
-        for w in iowaiters:
-            w[2].append(tuple(rdylist))
-            w[3].set()
-        iowaiters = []
-        timeout = 300
-        while ioqueue:
-            workitem = ioqueue.popleft()
-            # structure is function, args, list to append to ,event to set
-            if isinstance(workitem[1], tuple):  # positional arguments
-                try:
-                    workitem[2].append(workitem[0](*workitem[1]))
-                except Exception:
-                    traceback.print_exc()
-                workitem[3].set()
-            elif isinstance(workitem[1], dict):
-                try:
-                    workitem[2].append(workitem[0](**workitem[1]))
-                except Exception:
-                    traceback.print_exc()
-                workitem[3].set()
-            elif workitem[0] == 'wait':
-                if len(rdylist) > 0:
-                    workitem[2].append(tuple(rdylist))
-                    workitem[3].set()
-                else:
-                    ltimeout = workitem[1] - _monotonic_time()
-                    if ltimeout < timeout:
-                        timeout = ltimeout
-                    iowaiters.append(workitem)
+def define_worker():
+    class _IOWorker(threading.Thread):
+        def join(self):
+            Session._cleanup()
+            self.running = False
+            iosockets[0].sendto('\x01', ('::1', iosockets[0].getsockname()[1]))
+            super(_IOWorker, self).join()
+
+        def run(self):
+            self.running = True
+            global iothreadready
+            global selectdeadline
+            iowaiters = []
+            timeout = 300
+            iothreadready = True
+            while iothreadwaiters:
+                waiter = iothreadwaiters.pop()
+                waiter.set()
+            while self.running:
+                if timeout < 0:
+                    timeout = 0
+                selectdeadline = _monotonic_time() + timeout
+                tmplist, _, _ = select.select(iosockets, (), (), timeout)
+                # pessimistically move out the deadline
+                # doing it this early (before ioqueue is evaluated)
+                # this avoids other threads making a bad assumption
+                # about not having to break into the select
+                selectdeadline = _monotonic_time() + 300
+                _io_graball(iosockets)
+                for w in iowaiters:
+                    w[3].set()
+                iowaiters = []
+                timeout = 300
+                while ioqueue:
+                    workitem = ioqueue.popleft()
+                    # order: function, args, list to append to , event to set
+                    if isinstance(workitem[1], tuple):  # positional arguments
+                        try:
+                            workitem[2].append(workitem[0](*workitem[1]))
+                        except Exception:
+                            traceback.print_exc()
+                        workitem[3].set()
+                    elif isinstance(workitem[1], dict):
+                        try:
+                            workitem[2].append(workitem[0](**workitem[1]))
+                        except Exception:
+                            traceback.print_exc()
+                        workitem[3].set()
+                    elif workitem[0] == 'wait':
+                        if pktqueue:
+                            workitem[3].set()
+                        else:
+                            ltimeout = workitem[1] - _monotonic_time()
+                            if ltimeout < timeout:
+                                timeout = ltimeout
+                            iowaiters.append(workitem)
+    return _IOWorker
 
 
-def _io_apply(function, args):
-    global selectbreak
+pktqueue = collections.deque([])
+
+
+def _io_wait(timeout):
     evt = threading.Event()
     result = []
-    ioqueue.append((function, args, result, evt))
-    if not (function == 'wait' and selectdeadline < args):
-        os.write(selectbreak[1], '1')
+    deadline = timeout + _monotonic_time()
+    ioqueue.append(('wait', deadline, result, evt))
+    # Unfortunately, at least with eventlet patched threading, the wait()
+    # is a somewhat busy wait if given a deadline.  Workaround by having
+    # it piggy back on the select() in the io thread, which is a truly
+    # lazy wait even with eventlet involvement
+    if deadline < selectdeadline:
+        iosockets[0].sendto('\x01', ('::1', iosockets[0].getsockname()[1]))
     evt.wait()
-    return result[0]
 
 
 def _io_sendto(mysocket, packet, sockaddr):
@@ -152,14 +141,42 @@ def _io_sendto(mysocket, packet, sockaddr):
         pass
 
 
+def _io_graball(mysockets):
+        for mysocket in mysockets:
+            while True:
+                rdata = _io_recvfrom(mysocket, 3000)
+                if rdata is None:
+                    break
+                # If the payload is shorter than 4 bytes, it cannot
+                # be a useful packet.  Skip it entirely.
+                # This applies to the packet sent to self to break
+                # into the select
+                if len(rdata[0]) < 4:
+                    continue
+                rdata = rdata + (mysocket,)
+                pktqueue.append(rdata)
+
+
 def _io_recvfrom(mysocket, size):
     mysocket.setblocking(0)
-    ignoresockets.discard(mysocket)
     try:
         return mysocket.recvfrom(size)
     except socket.error:
         return None
 
+wintime = None
+try:
+    wintime = ctypes.windll.kernel32.GetTickCount64
+except AttributeError:
+    pass
+
+try:
+    IPPROTO_IPV6 = socket.IPPROTO_IPV6
+except AttributeError:
+    IPPROTO_IPV6 = 41  # This is the Win32 version of IPPROTO_IPV6, the only
+    # platform where python *doesn't* have this in socket that pyghmi is
+    # targetting.
+
 
 def _monotonic_time():
     """Provides a monotonic timer
@@ -170,13 +187,16 @@ def _monotonic_time():
     # Python does not provide one until 3.3, so we make do
     # for most OSes, os.times()[4] works well.
     # for microsoft, GetTickCount64
+    if wintime:
+        return wintime() / 1000.0
     return os.times()[4]
 
 
 def _poller(timeout=0):
-    if ignoresockets:
+    if pktqueue:
         return True
-    return _io_apply('wait', timeout + _monotonic_time())
+    _io_wait(timeout)
+    return pktqueue
 
 
 def _aespad(data):
@@ -259,47 +279,48 @@ class Session(object):
 
     @classmethod
     def _cleanup(cls):
-        global running
         for session in cls.bmc_handlers.itervalues():
             session.cleaningup = True
             session.logout()
-        running = False
 
     @classmethod
-    def _initsessions(cls):
-        atexit.register(cls._cleanup)
-
-    def _assignsocket(self):
+    def _assignsocket(cls, server=None):
         global iothread
         global iothreadready
         global iosockets
+        # seek for the least used socket.  As sessions close, they may free
+        # up slots in seemingly 'full' sockets.  This scheme allows those
+        # slots to be recycled
+        sorted_candidates = None
+        if server is None:
+            sorted_candidates = sorted(cls.socketpool.iteritems(),
+                                       key=operator.itemgetter(1))
+        if sorted_candidates and sorted_candidates[0][1] < MAX_BMCS_PER_SOCKET:
+            cls.socketpool[sorted_candidates[0][0]] += 1
+            return sorted_candidates[0][0]
+        # we need a new socket
+        tmpsocket = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)  # INET6
+                                    # can do IPv4 if you are nice to it
+        tmpsocket.setsockopt(IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
+        # Rather than wait until send() to bind, bind now so that we have
+        # a port number allocated no matter what
+        tmpsocket.bind(('', 0))
+        if server is None:
+            cls.socketpool[tmpsocket] = 1
+        else:
+            tmpsocket.bind(server)
+        iosockets.append(tmpsocket)
         if iothread is None:
-            self._initsessions()
             initevt = threading.Event()
             iothreadwaiters.append(initevt)
-            iothread = threading.Thread(target=_ioworker)
-            iothread.daemon = True
+            _IOWorker = define_worker()
+            iothread = _IOWorker()
             iothread.start()
             initevt.wait()
         elif not iothreadready:
             initevt = threading.Event()
             iothreadwaiters.append(initevt)
             initevt.wait()
-        # seek for the least used socket.  As sessions close, they may free
-        # up slots in seemingly 'full' sockets.  This scheme allows those
-        # slots to be recycled
-        sorted_candidates = sorted(self.socketpool.iteritems(),
-                                   key=operator.itemgetter(1))
-        if sorted_candidates and sorted_candidates[0][1] < MAX_BMCS_PER_SOCKET:
-            self.socketpool[sorted_candidates[0][0]] += 1
-            return sorted_candidates[0][0]
-        # we need a new socket
-        tmpsocket = _io_apply(socket.socket,
-                              (socket.AF_INET6, socket.SOCK_DGRAM))  # INET6
-                                    # can do IPv4 if you are nice to it
-        tmpsocket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
-        self.socketpool[tmpsocket] = 1
-        iosockets.append(tmpsocket)
         return tmpsocket
 
     def _sync_login(self, response):
@@ -326,7 +347,7 @@ class Session(object):
                 self = cls.bmc_handlers[sockaddr]
                 if (self.bmc == bmc and self.userid == userid and
                         self.password == password and self.kgo == kg and
-                        self.logged):
+                        (self.logged or self.logging)):
                     trueself = self
                 else:
                     del cls.bmc_handlers[sockaddr]
@@ -356,6 +377,7 @@ class Session(object):
         self.maxtimeout = 3  # be aggressive about giving up on initial packet
         self.incommand = False
         self.nameonly = 16  # default to name only lookups in RAKP exchange
+        self.servermode = False
         self.initialized = True
         self.cleaningup = False
         self.lastpayload = None
@@ -393,7 +415,7 @@ class Session(object):
             self.socket = self._assignsocket()
         self.login()
         if not self.async:
-            while not self.logged:
+            while self.logging and not self.logged:
                 Session.wait_for_rsp()
 
     def _mark_broken(self):
@@ -401,6 +423,7 @@ class Session(object):
         # deregister our keepalive facility
         Session.keepalive_sessions.pop(self, None)
         Session.waiting_sessions.pop(self, None)
+        self.logging = False
         if self.logged:
             self.logged = 0  # mark session as busted
             self._customkeepalives = None
@@ -436,6 +459,7 @@ class Session(object):
         self.integrityalgo = 0
         self.k1 = None
         self.rmcptag = 1
+        self.lastpayload = None
         self.ipmicallback = None
         self.sessioncontext = None
         self.sequencenumber = 0
@@ -449,6 +473,7 @@ class Session(object):
         #                 do not forsee a reason to adjust
         self.rqaddr = 0x81
 
+        self.logging = True
         self.logged = 0
         # NOTE(jbjohnso): when we confirm a working sockaddr, put it here to
         #                 skip getaddrinfo
@@ -507,7 +532,8 @@ class Session(object):
                                    # we are always the requestor for now
         seqincrement = 7  # IPMI spec forbids gaps bigger then 7 in seq number.
                        # Risk the taboo rather than violate the rules
-        while ((netfn, command, self.seqlun) in self.tabooseq and
+        while (not self.servermode and
+                (netfn, command, self.seqlun) in self.tabooseq and
                self.tabooseq[(netfn, command, self.seqlun)] and seqincrement):
             self.tabooseq[(self.expectednetfn, command, self.seqlun)] -= 1
                      # Allow taboo to eventually expire after a few rounds
@@ -526,7 +552,8 @@ class Session(object):
         else:
             rqaddr = self.rqaddr
             rsaddr = constants.IPMI_BMC_ADDRESS
-
+        if self.servermode:
+            rsaddr = self.clientaddr
         #figure 13-4, first two bytes are rsaddr and
         # netfn, for non-bridge request, rsaddr is always 0x20 since we are
         # addressing BMC while rsaddr is specified forbridge request
@@ -542,7 +569,8 @@ class Session(object):
             tail_csum = _checksum(*payload[3:])
             payload.append(tail_csum)
 
-        self._add_request_entry((self.expectednetfn, self.seqlun, command))
+        if not self.servermode:
+            self._add_request_entry((self.expectednetfn, self.seqlun, command))
         return payload
 
     def _generic_callback(self, response):
@@ -602,8 +630,17 @@ class Session(object):
         self.incommand = False
         return lastresponse
 
-    def _send_ipmi_net_payload(self, netfn, command, data, bridge_request=None,
-                               retry=True, delay_xmit=None):
+    def _send_ipmi_net_payload(self, netfn=None, command=None, data=[], code=0,
+                               bridge_request=None,
+                               retry=None, delay_xmit=None):
+        if retry is None:
+            retry = not self.servermode
+        if self.servermode:
+            data = [code] + data
+            if netfn is None:
+                netfn = self.clientnetfn
+            if command is None:
+                command = self.clientcommand
         ipmipayload = self._make_ipmi_payload(netfn, command, bridge_request,
                                               data)
         payload_type = constants.payload_types['ipmi']
@@ -805,28 +842,25 @@ class Session(object):
         self._req_priv_level()
 
     def _req_priv_level(self):
-        self.ipmicallback = self._got_priv_level
-        self._send_ipmi_net_payload(netfn=0x6,
-                                    command=0x3b,
+        self.logged = 1
+        response = self.raw_command(netfn=0x6, command=0x3b,
                                     data=[self.privlevel])
-
-    def _got_priv_level(self, response):
         if response['code']:
             if response['code'] in (0x80, 0x81) and self.privlevel == 4:
                 # some implementations will let us get this far,
                 # but suddenly get skiddish.  Try again in such a case
                 self.privlevel = 3
-                self.logged = 1
-                self.logout()
-                self._relog()
-                return
-            mysuffix = " while requesting privelege level %d for %s" % (
-                self.privlevel, self.userid)
-            errstr = get_ipmi_error(response, suffix=mysuffix)
-            if errstr:
-                self.onlogon({'error': errstr})
-                return
-        self.logged = 1
+                response = self.raw_command(netfn=0x6, command=0x3b,
+                                            data=[self.privlevel])
+            if response['code']:
+                self.logged = 0
+                mysuffix = " while requesting privelege level %d for %s" % (
+                    self.privlevel, self.userid)
+                errstr = get_ipmi_error(response, suffix=mysuffix)
+                if errstr:
+                    self.onlogon({'error': errstr})
+                    return
+        self.logging = False
         Session.keepalive_sessions[self] = {}
         Session.keepalive_sessions[self]['ipmisession'] = self
         Session.keepalive_sessions[self]['timeout'] = _monotonic_time() + \
@@ -882,15 +916,6 @@ class Session(object):
         self._initsession()
         self._get_channel_auth_cap()
 
-    @classmethod
-    def pulltoqueue(cls, mysockets, queue):
-        for mysocket in mysockets:
-            while True:
-                rdata = _io_apply(_io_recvfrom, (mysocket, 3000))
-                if rdata is None:
-                    break
-                queue.append(rdata)
-
     @classmethod
     def wait_for_rsp(cls, timeout=None, callout=True):
         """IPMI Session Event loop iteration
@@ -950,12 +975,9 @@ class Session(object):
         if timeout is None:
             return 0
         if _poller(timeout=timeout):
-            pktqueue = collections.deque([])
-            cls.pulltoqueue(iosockets, pktqueue)
-            while len(pktqueue):
-                (data, sockaddr) = pktqueue.popleft()
-                cls._route_ipmiresponse(sockaddr, data)
-                cls.pulltoqueue(iosockets, pktqueue)
+            while pktqueue:
+                (data, sockaddr, mysocket) = pktqueue.popleft()
+                cls._route_ipmiresponse(sockaddr, data, mysocket)
         sessionstodel = []
         sessionstokeepalive = []
         for session, parms in cls.keepalive_sessions.iteritems():
@@ -977,6 +999,9 @@ class Session(object):
                                               # to avoid confusing the for loop
         for session in sessionstodel:
             cls.waiting_sessions.pop(session, None)
+        # one loop iteration to make sure recursion doesn't induce redundant
+        # timeouts
+        for session in sessionstodel:
             session._timedout()
         return len(cls.waiting_sessions)
 
@@ -1036,14 +1061,16 @@ class Session(object):
             self._mark_broken()
 
     @classmethod
-    def _route_ipmiresponse(cls, sockaddr, data):
+    def _route_ipmiresponse(cls, sockaddr, data, mysocket):
         if not (data[0] == '\x06' and data[2:4] == '\xff\x07'):  # not ipmi
             return
         try:
             cls.bmc_handlers[sockaddr]._handle_ipmi_packet(data,
                                                            sockaddr=sockaddr)
         except KeyError:
-            pass
+            # check if we have a server attached to the target socket
+            if mysocket in cls.bmc_handlers:
+                cls.bmc_handlers[mysocket].sessionless_data(data, sockaddr)
 
     def _handle_ipmi_packet(self, data, sockaddr=None):
         if self.sockaddr is None and sockaddr is not None:
@@ -1089,16 +1116,33 @@ class Session(object):
         else:
             return  # unrecognized data, assume evil
 
+    def _got_rakp1(self, data):
+        # stub, client sessions ignore rakp2
+        pass
+
+    def _got_rakp3(self, data):
+        #stub, client sessions ignore rakp3
+        pass
+
+    def _got_rmcp_openrequest(self, data):
+        pass
+
     def _handle_ipmi2_packet(self, rawdata):
         data = list(struct.unpack("%dB" % len(rawdata), rawdata))
                     #now need mutable array
         ptype = data[5] & 0b00111111
         # the first 16 bytes are header information as can be seen in 13-8 that
         # we will toss out
-        if ptype == 0x11:  # rmcp+ response
+        if ptype == 0x10:
+            return self._got_rmcp_openrequest(data[16:])
+        elif ptype == 0x11:  # rmcp+ response
             return self._got_rmcp_response(data[16:])
+        elif ptype == 0x12:
+            return self._got_rakp1(data[16:])
         elif ptype == 0x13:
             return self._got_rakp2(data[16:])
+        elif ptype == 0x14:
+            return self._got_rakp3(data[16:])
         elif ptype == 0x15:
             return self._got_rakp4(data[16:])
         elif ptype == 0 or ptype == 1:  # good old ipmi payload or sol
@@ -1236,7 +1280,8 @@ class Session(object):
             struct.pack("2B", self.nameonly | self.privlevel, userlen) +\
             self.userid
         expectedhash = hmac.new(self.password, hmacdata, hashlib.sha1).digest()
-        givenhash = struct.pack("%dB" % len(data[40:]), *data[40:])
+        hashlen = len(expectedhash)
+        givenhash = struct.pack("%dB" % hashlen, *data[40:hashlen + 40])
         if givenhash != expectedhash:
             self.sessioncontext = "FAILED"
             self.onlogon({'error': "Incorrect password provided"})
@@ -1265,6 +1310,7 @@ class Session(object):
             struct.pack("2B", self.nameonly | self.privlevel,
                         len(self.userid)) +\
             self.userid
+
         authcode = hmac.new(self.password, hmacdata, hashlib.sha1).digest()
         payload += list(struct.unpack("%dB" % len(authcode), authcode))
         self.send_payload(
@@ -1301,7 +1347,8 @@ class Session(object):
             self.remoteguid
         expectedauthcode = hmac.new(self.sik, hmacdata,
                                     hashlib.sha1).digest()[:12]
-        authcode = struct.pack("%dB" % len(data[8:]), *data[8:])
+        aclen = len(expectedauthcode)
+        authcode = struct.pack("%dB" % aclen, *data[8:aclen + 8])
         if authcode != expectedauthcode:
             self.onlogon({'error': "Invalid RAKP4 integrity code (wrong Kg?)"})
             return
@@ -1318,6 +1365,13 @@ class Session(object):
         # For now, skip the checksums since we are in LAN only,
         # TODO(jbjohnso): if implementing other channels, add checksum checks
         # here
+        if self.servermode:
+            self.seqlun = payload[4]
+            self.clientaddr = payload[3]
+            self.clientnetfn = (payload[1] >> 2) + 1
+            self.clientcommand = payload[5]
+            self._parse_payload(payload)
+            return
         entry = (payload[1] >> 2, payload[4], payload[5])
         if self._lookup_request_entry(entry):
             self._remove_request_entry(entry)
@@ -1343,7 +1397,6 @@ class Session(object):
             return -1
 
     def _parse_payload(self, payload):
-
         if hasattr(self, 'hasretried') and self.hasretried:
             self.hasretried = 0
             self.tabooseq[
@@ -1353,8 +1406,9 @@ class Session(object):
         self.expectednetfn = 0x1ff  # bigger than one byte means it can never
                                     # match the one byte value by mistake
         self.expectedcmd = 0x1ff
-        self.seqlun += 4  # prepare seqlun for next transmit
-        self.seqlun &= 0xff  # when overflowing, wrap around
+        if not self.servermode:
+            self.seqlun += 4  # prepare seqlun for next transmit
+            self.seqlun &= 0xff  # when overflowing, wrap around
         Session.waiting_sessions.pop(self, None)
         self.lastpayload = None  # render retry mechanism utterly incapable of
                                  # doing anything, though it shouldn't matter
@@ -1365,11 +1419,15 @@ class Session(object):
         # ^^ remove header of rsaddr/netfn/lun/checksum/rq/seq/lun
         del payload[-1]  # remove the trailing checksum
         response['command'] = payload[0]
-        response['code'] = payload[1]
-        del payload[0:2]
-        response['data'] = payload
+        if self.servermode:
+            del payload[0:1]
+            response['data'] = payload
+        else:
+            response['code'] = payload[1]
+            del payload[0:2]
+            response['data'] = payload
         self.timeout = initialtimeout + (0.5 * random.random())
-        if len(self.pendingpayloads) > 0:
+        if not self.servermode and len(self.pendingpayloads) > 0:
             (nextpayload, nextpayloadtype, retry) = \
                 self.pendingpayloads.popleft()
             self.send_payload(payload=nextpayload,
@@ -1383,7 +1441,7 @@ class Session(object):
         self.nowait = True
         self.timeout += 1
         if self.timeout > self.maxtimeout:
-            response = {'error': 'timeout'}
+            response = {'error': 'timeout', 'code': 0xffff}
             self.ipmicallback(response)
             self.nowait = False
             self._mark_broken()
@@ -1421,18 +1479,14 @@ class Session(object):
         if self.sequencenumber:  # seq number of zero will be left alone, it is
                                 # special, otherwise increment
             self.sequencenumber += 1
-        if retry:
+        if delay_xmit is not None:
             Session.waiting_sessions[self] = {}
             Session.waiting_sessions[self]['ipmisession'] = self
-            Session.waiting_sessions[self]['timeout'] = self.timeout + \
-                _monotonic_time()
-        if delay_xmit is not None:
             Session.waiting_sessions[self]['timeout'] = delay_xmit + \
                 _monotonic_time()
             return  # skip transmit, let retry timer do it's thing
         if self.sockaddr:
-            _io_apply(_io_sendto,
-                      (self.socket, self.netpacket, self.sockaddr))
+            _io_sendto(self.socket, self.netpacket, self.sockaddr)
         else:  # he have not yet picked a working sockaddr for this connection,
               # try all the candidates that getaddrinfo provides
             self.allsockaddrs = []
@@ -1448,11 +1502,15 @@ class Session(object):
                         sockaddr = (newhost, sockaddr[1], 0, 0)
                     self.allsockaddrs.append(sockaddr)
                     Session.bmc_handlers[sockaddr] = self
-                    _io_apply(_io_sendto, (self.socket,
-                              self.netpacket, sockaddr))
+                    _io_sendto(self.socket, self.netpacket, sockaddr)
             except socket.gaierror:
                 raise exc.IpmiException(
                     "Unable to transmit to specified address")
+        if retry:
+            Session.waiting_sessions[self] = {}
+            Session.waiting_sessions[self]['ipmisession'] = self
+            Session.waiting_sessions[self]['timeout'] = self.timeout + \
+                _monotonic_time()
 
     def logout(self):
         if not self.logged:
@@ -1467,6 +1525,7 @@ class Session(object):
         # stop trying for a keepalive,
         Session.keepalive_sessions.pop(self, None)
         self.logged = 0
+        self.logging = False
         self._customkeepalives = None
         self.nowait = False
         self.socketpool[self.socket] -= 1
