@@ -1,7 +1,8 @@
 # vim: tabstop=4 shiftwidth=4 softtabstop=4
 # coding=utf8
 
-# Copyright 2013 IBM Corporation
+# Copyright 2014 IBM Corporation
+# Copyright 2015 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -185,6 +186,7 @@ class SensorReading(object):
     """
 
     def __init__(self, reading, suffix):
+        self.broken_sensor_ids = {}
         self.health = const.Health.Ok
         self.type = reading['type']
         self.value = None
@@ -193,6 +195,7 @@ class SensorReading(object):
         try:
             self.health = reading['health']
             self.states = reading['states']
+            self.state_ids = reading['state_ids']
             self.value = reading['value']
             self.imprecision = reading['imprecision']
         except KeyError:
@@ -206,6 +209,7 @@ class SensorReading(object):
         return repr({
             'value': self.value,
             'states': self.states,
+            'state_ids': self.state_ids,
             'units': self.units,
             'imprecision': self.imprecision,
             'name': self.name,
@@ -243,10 +247,11 @@ class SDREntry(object):
     external code to pay attention to this class.
     """
 
-    def __init__(self, entrybytes, reportunsupported=False):
+    def __init__(self, entrybytes, ipmicmd, reportunsupported=False):
         # ignore record id for now, we only care about the sensor number for
         # moment
         self.reportunsupported = reportunsupported
+        self.ipmicmd = ipmicmd
         if entrybytes[2] != 0x51:
             # only recognize '1.5', the only version defined at time of writing
             raise NotImplementedError
@@ -268,9 +273,6 @@ class SDREntry(object):
             self.sdrtype = TYPE_UNKNOWN   # assume undefined
             self.oem_decode(entrybytes[5:])
         elif self.reportunsupported:
-            #will remove once I see it stop being thrown for now
-            #perhaps need some explicit mode to check for
-            #unsupported things, but make do otherwise
             raise NotImplementedError
         else:
             self.sdrtype = TYPE_UNKNOWN
@@ -300,6 +302,9 @@ class SDREntry(object):
         self.sdrtype = TYPE_FRU
         self.fru_name = self.tlv_decode(entry[10], entry[11:])
         self.fru_number = entry[1]
+        self.fru_logical = (entry[2] & 0b10000000) == 0b10000000
+        # 0x8  to 0x10..  0 unspecified except on 0x10, 1 is dimm
+        self.fru_type_and_modifier = (entry[5] << 8) + entry[6]
 
     def association_decode(self, entry):
         # table 43-4 Entity Associaition Record
@@ -311,12 +316,17 @@ class SDREntry(object):
         self._common_decode(entry)
         self.sensor_name = self.tlv_decode(entry[26], entry[27:])
 
+    def assert_trap_value(self, offset):
+        trapval = (self.sensor_type_number << 16) + (self.reading_type << 8)
+        return trapval + offset
+
     def _common_decode(self, entry):
         # compact and full are very similar
         # this function handles the common aspects of compact and full
         # offsets from spec, minus 6
         self.sensor_number = entry[2]
-        self.entity = ipmiconst.entity_ids[entry[3]]
+        self.entity = ipmiconst.entity_ids.get(
+            entry[3], 'Unknown entity {0}'.format(entry[3]))
         self.sensor_type_number = entry[7]
         try:
             self.sensor_type = ipmiconst.sensor_type_codes[entry[7]]
@@ -327,10 +337,19 @@ class SDREntry(object):
             # 1: generic threshold based
             # 0x6f: discrete sensor-specific from table 42-3, sensor offsets
             # all others per table 42-2, generic discrete
-        self.numeric_format = (entry[15] & 0b11000000) >> 6
-        # the spec technically reserves numeric_format for compact sensor
-        # numeric, but common treatment won't break currently
+        # numeric format is one of:
         # 0 - unsigned, 1 - 1s complement, 2 - 2s complement, 3 - ignore number
+        # compact records are supposed to always write it as '3', presumably
+        # to allow for the concept of a compact record with a numeric format
+        # even though numerics are not allowed today.  Some implementations
+        # violate the spec and do something other than 3 today.  Tolerate
+        # the violation under the assumption that things are not so hard up
+        # that there will ever be a need for compact sensors supporting numeric
+        # values
+        if self.rectype == 2:
+            self.numeric_format = 3
+        else:
+            self.numeric_format = (entry[15] & 0b11000000) >> 6
         self.sensor_rate = sensor_rates[(entry[15] & 0b111000) >> 3]
         self.unit_mod = ""
         if (entry[15] & 0b110) == 0b10:  # unit1 by unit2
@@ -364,10 +383,10 @@ class SDREntry(object):
             # factors on the fly.
             # the formula could apply if we bother with nominal
             # reading interpretation
-            self.decode_formula(entry)
+            self.decode_formula(entry[19:25])
 
     def _decode_state(self, state):
-        mapping = ipmiconst.discrete_type_offsets
+        mapping = ipmiconst.generic_type_offsets
         try:
             if self.reading_type in mapping:
                 desc = mapping[self.reading_type][state]['desc']
@@ -383,13 +402,14 @@ class SDREntry(object):
             desc = "Unknown state %d for reading type %d/sensor type %d" % (
                 state, self.reading_type, self.sensor_type_number)
             health = const.Health.Warning
-        return (desc, health)
+        return desc, health
 
     def decode_sensor_reading(self, reading):
         numeric = None
         output = {
             'name': self.sensor_name,
             'type': self.sensor_type,
+            'id': self.sensor_number,
         }
         if reading[1] & 0b100000:
             output['unavailable'] = 1
@@ -417,6 +437,7 @@ class SDREntry(object):
             upper = 'lower'
             lower = 'upper'
         output['states'] = []
+        output['state_ids'] = []
         output['health'] = const.Health.Ok
         if discrete:
             for state in range(8):
@@ -424,99 +445,120 @@ class SDREntry(object):
                     statedesc, health = self._decode_state(state)
                     output['health'] |= health
                     output['states'].append(statedesc)
+                    output['state_ids'].append(self.assert_trap_value(state))
             if len(reading) > 3:
                 for state in range(7):
                     if reading[3] & (0b1 << state):
                         statedesc, health = self._decode_state(state + 8)
                         output['health'] |= health
                         output['states'].append(statedesc)
+                        output['state_ids'].append(
+                            self.assert_trap_value(state + 8))
         else:
             if reading[2] & 0b1:
                 output['health'] |= const.Health.Warning
                 output['states'].append(lower + " non-critical threshold")
+                output['state_ids'].append(self.assert_trap_value(1))
             if reading[2] & 0b10:
                 output['health'] |= const.Health.Critical
                 output['states'].append(lower + " critical threshold")
+                output['state_ids'].append(self.assert_trap_value(2))
             if reading[2] & 0b100:
                 output['health'] |= const.Health.Failed
                 output['states'].append(lower + " non-recoverable threshold")
+                output['state_ids'].append(self.assert_trap_value(3))
             if reading[2] & 0b1000:
                 output['health'] |= const.Health.Warning
                 output['states'].append(upper + " non-critical threshold")
+                output['state_ids'].append(self.assert_trap_value(4))
             if reading[2] & 0b10000:
                 output['health'] |= const.Health.Critical
                 output['states'].append(upper + " critical threshold")
+                output['state_ids'].append(self.assert_trap_value(5))
             if reading[2] & 0b100000:
                 output['health'] |= const.Health.Failed
                 output['states'].append(upper + " non-recoverable threshold")
+                output['state_ids'].append(self.assert_trap_value(6))
         return SensorReading(output, self.unit_suffix)
 
+    def _set_tmp_formula(self, value):
+        rsp = self.ipmicmd.raw_command(netfn=4, command=0x23,
+                                       data=(self.sensor_number, value))
+        # skip next reading field, not used in on-demand situation
+        self.decode_formula(rsp['data'][1:])
+
     def decode_value(self, value):
         # Take the input value and return meaningful value
-        if self.linearization == 0x70:  # direct calling code to get factors
-            #TODO(jbjohnso): implement get sensor reading factors support for
-            #non linear sensor
-            raise NotImplementedError("Need to do get sensor reading factors")
+        linearization = self.linearization
+        if linearization > 11:  # direct calling code to get factors
+            # for now, we will get the factors on demand
+            # the facility is engineered such that at construction
+            # time the entire BMC table should be fetchable in a reasonable
+            # fashion.  However for now opt for retrieving rows as needed
+            # rather than tracking all that information for a relatively
+            # rare behavior
+            self._set_tmp_formula(value)
+            linearization = 0
         # time to compute the pre-linearization value.
         decoded = float((value * self.m + self.b) *
                         (10 ** self.resultexponent))
-        if self.linearization == 0:
+        if linearization == 0:
             return decoded
-        elif self.linearization == 1:
+        elif linearization == 1:
             return math.log(decoded)
-        elif self.linearization == 2:
+        elif linearization == 2:
             return math.log(decoded, 10)
-        elif self.linearization == 3:
+        elif linearization == 3:
             return math.log(decoded, 2)
-        elif self.linearization == 4:
+        elif linearization == 4:
             return math.exp(decoded)
-        elif self.linearization == 5:
+        elif linearization == 5:
             return 10 ** decoded
-        elif self.linearization == 6:
+        elif linearization == 6:
             return 2 ** decoded
-        elif self.linearization == 7:
+        elif linearization == 7:
             return 1 / decoded
-        elif self.linearization == 8:
+        elif linearization == 8:
             return decoded ** 2
-        elif self.linearization == 9:
+        elif linearization == 9:
             return decoded ** 3
-        elif self.linearization == 10:
+        elif linearization == 10:
             return math.sqrt(decoded)
-        elif self.linearization == 11:
+        elif linearization == 11:
             return decoded ** (1.0/3)
         else:
             raise NotImplementedError
 
     def decode_formula(self, entry):
         self.m = \
-            twos_complement(entry[19] + ((entry[20] & 0b11000000) << 2), 10)
-        self.tolerance = entry[20] & 0b111111
+            twos_complement(entry[0] + ((entry[1] & 0b11000000) << 2), 10)
+        self.tolerance = entry[1] & 0b111111
         self.b = \
-            twos_complement(entry[21] + ((entry[22] & 0b11000000) << 2), 10)
-        self.accuracy = (entry[22] & 0b111111) + \
-            (entry[23] & 0b11110000) << 2
-        self.accuracyexp = (entry[23] & 0b1100) >> 2
-        self.direction = entry[23] & 0b11
+            twos_complement(entry[2] + ((entry[3] & 0b11000000) << 2), 10)
+        self.accuracy = (entry[3] & 0b111111) + \
+            (entry[4] & 0b11110000) << 2
+        self.accuracyexp = (entry[4] & 0b1100) >> 2
+        self.direction = entry[4] & 0b11
             #0 = n/a, 1 = input, 2 = output
-        self.resultexponent = twos_complement((entry[24] & 0b11110000) >> 4, 4)
-        bexponent = twos_complement(entry[24] & 0b1111, 4)
+        self.resultexponent = twos_complement((entry[5] & 0b11110000) >> 4, 4)
+        bexponent = twos_complement(entry[5] & 0b1111, 4)
         # might as well do the math to 'b' now rather than wait for later
         self.b = self.b * (10**bexponent)
 
     def tlv_decode(self, tlv, data):
         # Per IPMI 'type/length byte format
-        type = (tlv & 0b11000000) >> 6
+        ipmitype = (tlv & 0b11000000) >> 6
         if not len(data):
             return ""
-        if type == 0:  # Unicode per 43.15 in ipmi 2.0 spec
+        if ipmitype == 0:  # Unicode per 43.15 in ipmi 2.0 spec
             # the spec is not specific about encoding, assuming utf8
             return unicode(struct.pack("%dB" % len(data), *data), "utf_8")
-        elif type == 1:  # BCD '+'
+        elif ipmitype == 1:  # BCD '+'
             tmpl = "%02X" * len(data)
             tstr = tmpl % tuple(data)
             tstr = tstr.replace("A", " ").replace("B", "-").replace("C", ".")
             return tstr.replace("D", ":").replace("E", ",").replace("F", "_")
-        elif type == 2:  # 6 bit ascii, start at 0x20 and stop when out of bits
+        elif ipmitype == 2:  # 6 bit ascii, start at 0x20
             # the ordering is very peculiar and is best understood from
             # IPMI SPEC "6-bit packed ascii example
             tstr = ""
@@ -528,7 +570,7 @@ class SDREntry(object):
                             (data[1] >> 4) + 0x20)
                 tstr += chr((data[2] >> 2) + 0x20)
             return tstr
-        elif type == 3:  # ACSII+LATIN1
+        elif ipmitype == 3:  # ACSII+LATIN1
             return struct.pack("%dB" % len(data), *data)
 
 
@@ -569,9 +611,15 @@ class SDR(object):
             self.aux_fw = self.decode_aux(rsp['data'][11:15])
         self.get_sdr()
 
+    def get_sdr_reservation(self):
+        rsp = self.ipmicmd.raw_command(netfn=0x0a, command=0x22)
+        if rsp['code'] != 0:
+            raise exc.IpmiException(rsp['error'])
+        return rsp['data'][0] + (rsp['data'][1] << 8)
+
     def get_sdr(self):
-        rsp = self.ipmicmd.raw_command(netfn=0x0a, command=0x20)
-        if (rsp['data'][0] != 0x51):
+        repinfo = self.ipmicmd.raw_command(netfn=0x0a, command=0x20)
+        if (repinfo['data'][0] != 0x51):
             # we only understand SDR version 51h, the only version defined
             # at time of this writing
             raise NotImplementedError
@@ -589,32 +637,77 @@ class SDR(object):
         rsvid = 0  # partial 'get sdr' will require this
         offset = 0
         size = 0xff
+        chunksize = 128
+        self.broken_sensor_ids = {}
         while recid != 0xffff:  # per 33.12 Get SDR command, 0xffff marks end
-            rqdata = [rsvid & 0xff, rsvid >> 8,
-                      recid & 0xff, recid >> 8,
-                      offset, size]
-            rsp = self.ipmicmd.raw_command(netfn=0x0a, command=0x23,
-                                           data=rqdata)
-            newrecid = (rsp['data'][1] << 8) + rsp['data'][0]
-            self.add_sdr(rsp['data'][2:])
+            newrecid = 0
+            currlen = 0
+            sdrdata = bytearray()
+            while True:  # loop until SDR fetched wholly
+                if size != 0xff and rsvid == 0:
+                    rsvid = self.get_sdr_reservation()
+                rqdata = [rsvid & 0xff, rsvid >> 8,
+                          recid & 0xff, recid >> 8,
+                          offset, size]
+                sdrrec = self.ipmicmd.raw_command(netfn=0x0a, command=0x23,
+                                                  data=rqdata)
+                if sdrrec['code'] == 0xca:
+                    if size == 0xff:  # get just 5 to get header to know length
+                        size = 5
+                    elif size > 5:
+                        size /= 2
+                        # push things over such that it's less
+                        # likely to be just 1 short of a read
+                        # and incur a whole new request
+                        size += 2
+                        chunksize = size
+                    continue
+                if sdrrec['code'] == 0xc5:  # need a new reservation id
+                    rsvid = 0
+                    continue
+                if sdrrec['code'] != 0:
+                    raise exc.IpmiException(sdrrec['error'])
+                if newrecid == 0:
+                    newrecid = (sdrrec['data'][1] << 8) + sdrrec['data'][0]
+                if currlen == 0:
+                    currlen = sdrrec['data'][6] + 5  # compensate for header
+                sdrdata.extend(sdrrec['data'][2:])
+                # determine next offset to use based on current offset and the
+                # size used last time.
+                offset += size
+                if offset >= currlen:
+                    break
+                if size == 5 and offset == 5:
+                    # bump up size after header retrieval
+                    size = chunksize
+                if (offset + size) > currlen:
+                    size = currlen - offset
+            self.add_sdr(sdrdata)
+            offset = 0
+            if size != 0xff:
+                size = 5
             if newrecid == recid:
                 raise exc.BmcErrorException("Incorrect SDR record id from BMC")
             recid = newrecid
+        for sid in self.broken_sensor_ids:
+            del self.sensors[sid]
 
     def get_sensor_numbers(self):
         return self.sensors.iterkeys()
 
     def add_sdr(self, sdrbytes):
-        newent = SDREntry(sdrbytes)
+        newent = SDREntry(sdrbytes, self.ipmicmd)
         if newent.sdrtype == TYPE_SENSOR:
             id = newent.sensor_number
             if id in self.sensors:
-                raise exc.BmcErrorException("Duplicate sensor number " + id)
+                self.broken_sensor_ids[id] = True
+                return
             self.sensors[id] = newent
         elif newent.sdrtype == TYPE_FRU:
             id = newent.fru_number
             if id in self.fru:
-                raise exc.BmcErrorException("Duplicate FRU identifier " + id)
+                self.broken_sensor_ids[id] = True
+                return
             self.fru[id] = newent
 
     def decode_aux(self, auxdata):
